from __future__ import annotations

from .StiBaseResult import StiBaseResult
from ..enums.StiFrameworkType import StiFrameworkType
import codecs
import json
import typing

if typing.TYPE_CHECKING:
    from .StiBaseHandler import StiBaseHandler

class StiBaseResponse:
    """
    The result of executing an event handler request. 
    You can get the data, its type and other parameters necessary to create a web server response.
    """

    handler: StiBaseHandler = None
    result: StiBaseResult = None


### Private

    def __getArrtibutes(obj: object):
        return { attr: getattr(obj, attr) for attr in dir(obj) if not callable(getattr(obj, attr)) and not attr.startswith('_') }


### Properties

    @property
    def origin(self) -> str:
        """Returns the detected origin url for the handler response. Can be used for the 'Access-Control-Allow-Origin' header of the response."""

        return self.handler.origin

    @property
    def mimetype(self) -> str:
        """Returns the mime-type for the handler response."""

        return 'application/json'
    
    @property
    def contentType(self) -> str:
        """Returns the content type for the handler response. Can be used for the 'Content-Type' header of the response."""

        return self.mimetype + '; charset=utf-8'
    
    @property
    def data(self) -> bytes:
        """Returns the handler response as a byte array. When using encryption, the response will be encrypted and encoded into a Base64 string."""

        data = json.dumps(self.result, default = StiBaseResponse.__getArrtibutes)
        if self.handler.request.encryptData == True:
            data = codecs.encode(codecs.encode(data.encode(), 'base64').decode(), 'rot13')
        return data.encode()

### Public

    def getFrameworkResponse(self):
        """
        Returns a response intended for one of the supported frameworks specified in the arguments. 
        The supported frameworks are in the 'StiFrameworkType' enum. 
        If the framework is not specified, it will be determined based on the request.
        """
        
        if self.handler.framework == StiFrameworkType.DJANGO:
            try:
                from django.http import HttpResponse
                contentType = self.contentType
                response = HttpResponse(self.data, content_type = contentType)
                return response
            except:
                return self

        if self.handler.framework == StiFrameworkType.FLASK:
            try:
                from flask import make_response
                response = make_response(self.data)
                response.mimetype = self.mimetype
                response.headers.add('Access-Control-Allow-Origin', self.origin)
                return response
            except:
                return self
            
        if self.handler.framework == StiFrameworkType.TORNADO:
            try:
                return self.data
            except:
                return self

        return self
    

### Constructor

    def __init__(self, handler: StiBaseHandler, result: StiBaseResult = None):
        self.handler = handler
        self.result = result if result != None else self.handler.getResult()



from .enums.StiDatabaseType import StiDatabaseType
from .classes.StiDataResult import StiDataResult
from .classes.StiConnectionInfo import StiConnectionInfo
from datetime import timedelta
from datetime import date
from datetime import time
from datetime import datetime
from decimal import Decimal
from uuid import UUID
import re
import codecs

class StiDataAdapter:
    version: str = ''
    checkVersion: bool = False

    connectionString: str = None
    connectionInfo: StiConnectionInfo = None
    connectionLink: object = None

    def getOdbcConnectionString(self):
        connectionString: str = \
            f'Driver={self.connectionInfo.driver};' \
            f'Server={self.connectionInfo.host};' \
            f'Database={self.connectionInfo.database};' \
            f'Uid={self.connectionInfo.userId};' \
            f'Pwd={self.connectionInfo.password};'
        
        return connectionString

    def connect(self):
        return StiDataResult.getSuccess(self)
    
    def connectOdbc(self):
        from .StiOdbcAdapter import StiOdbcAdapter
        adapter = StiOdbcAdapter()
        adapter.connectionString = self.getOdbcConnectionString()
        adapter.connectionInfo = self.connectionInfo

        result = adapter.connect()
        self.connectionLink = adapter.connectionLink
        return result

    def disconnect(self):
        if (not self.connectionLink is None):
            self.connectionLink.close()
        self.connectionLink = None

    def test(self):
        result = self.connect()
        if (result.success):
            self.disconnect()
        return result
    
    def process(self):
        self.connectionInfo = StiConnectionInfo()
        return True
    
    def parseParameters(self, parameterNames: dict[str, list[str]]):
        parameters = self.connectionString.split(';')

        for parameter in parameters:
            name = ''
            value = parameter
            if (parameter.find('=') >= 0):
                pos = parameter.find('=')
                name = parameter[0:pos].strip().lower()
                value = parameter[pos + 1:].strip()

            unknown = True
            if (len(parameterNames) > 0):
                for key, names in parameterNames.items():
                    if (name in names):
                        setattr(self.connectionInfo, key, value)
                        unknown = False
                        break

            if (unknown):
                self.parseUnknownParameter(parameter, name, value)

        return True
    
    def parseUnknownParameter(self, parameter: str, name: str, value: str):
        pass

    def parseType(self, meta: tuple):
        types = {
            'boolean': [bool],
            'int': [int],
            'number': [float, complex, Decimal],
            'time': [timedelta],
            'datetime': [time, date, datetime],
            'array': [bytes, bytearray],
            'string': [str]
        }

        for key, array in types.items():
            if (meta[1] in array):
                return key

        return 'string'
    
    def getValueType(self, value: object, index: int, types: list):
        return types[index]

    def getValue(self, value: object, valueType: str):
        if (value is None):
            return None

        if (valueType == 'array'):
            valueBytes = value.encode() if type(value) == str else value
            return codecs.encode(valueBytes, 'base64').decode()
        
        if (valueType == 'datetime'):
            valueDateTime: datetime = value
            return valueDateTime.strftime('%Y-%m-%dT%H:%M:%S.000')
        
        if (valueType == 'time'):
            valueDelta: timedelta = value
            hours = int(valueDelta.seconds / 3600) + valueDelta.days * 24
            minutes = int(valueDelta.seconds / 60) % 60
            seconds = int(valueDelta.seconds) % 60
            return '{:02d}:'.format(hours) + '{:02d}:'.format(minutes) + '{:02d}.000'.format(seconds)
        
        if (type(value) == Decimal):
            return float(value)
        
        if (type(value) == UUID):
            return str(value)
        
        return value
    
    def makeQuery(self, procedure: str, parameters: list):
        paramsString: str = ''
        for name in parameters:
            if (len(paramsString) > 0):
                paramsString += ', '
            paramsString += '@' + name

        return paramsString
    
    def executeQuery(self, queryString: str):
        result = self.connect()
        if (result.success):
            result.types = []
            result.columns = []
            result.rows = []

            result = self.executeNative(queryString, result)
            self.disconnect()

        return result
    
    def executeNative(self, queryString: str, result: StiDataResult):
        cursor = self.connectionLink.cursor()

        try:
            cursor.execute(queryString)
        except Exception as e:
            message = str(e)
            cursor.close()
            return StiDataResult.getError(self, message)
        
        for meta in cursor.description:
            result.columns.append(meta[0])
            columnType = self.parseType(meta)
            result.types.append(columnType)

        try:
            for rowItem in cursor:
                row = []
                for rowValue in rowItem:
                    valueType = self.getValueType(rowValue, len(row), result.types)
                    value = self.getValue(rowValue, valueType)
                    row.append(value)
                result.rows.append(row)
        except Exception as e:
            message = str(e)
            return StiDataResult.getError(self, message)
        finally:
            cursor.close()

        result.count = len(result.rows)
        return result

    def getDataAdapter(database: str, connectionString: str):
        if database == StiDatabaseType.MYSQL:
            from .StiMySqlAdapter import StiMySqlAdapter
            return StiMySqlAdapter(connectionString)
        
        elif database == StiDatabaseType.MSSQL:
            from .StiMsSqlAdapter import StiMsSqlAdapter
            return StiMsSqlAdapter(connectionString)
        
        elif database == StiDatabaseType.FIREBIRD:
            from .StiFirebirdAdapter import StiFirebirdAdapter
            return StiFirebirdAdapter(connectionString)
        
        elif database == StiDatabaseType.POSTGRESQL:
            from .StiPostgreSqlAdapter import StiPostgreSqlAdapter
            return StiPostgreSqlAdapter(connectionString)
        
        elif database == StiDatabaseType.ORACLE:
            from .StiOracleAdapter import StiOracleAdapter
            return StiOracleAdapter(connectionString)
        
        elif database == StiDatabaseType.ODBC:
            from .StiOdbcAdapter import StiOdbcAdapter
            return StiOdbcAdapter(connectionString)
        
        elif database == StiDatabaseType.MONGODB:
            from .StiMongoDbAdapter import StiMongoDbAdapter
            return StiMongoDbAdapter(connectionString)
        
        return None
    
    def applyQueryParameters(query: str, parameters: list, escape: bool):
        result: str = ''

        while (query.find('@') >= 0):
            result += query[0:query.find('@')]
            query = query[query.find('@') + 1:]

            parameterName: str = ''
            while (len(query) > 0):
                char = query[0];
                if (re.search('/[a-zA-Z0-9_-]/', char) is None):
                    break

                parameterName += char
                query = query[1:]

            replaced: bool = False
            '''foreach ($parameters as $key => $item) {
                if (strtolower($key) == strtolower($parameterName)) {
                    switch ($item->typeGroup) {
                        case 'number':
                            $result .= floatval($item->value);
                            break;

                        case 'datetime':
                            $result .= "'" . $item->value . "'";
                            break;

                        default:
                            $result .= "'" . ($escape ? addcslashes($item->value, "\\\"'") : $item->value) . "'";
                            break;
                    }

                    $replaced = true;
                }
            }'''

            if (replaced == False):
                result += '@' + parameterName

        return result + query
    
    def __init__(self, connectionString: str):
        self.connectionString = connectionString.strip()
        self.process()


from .StiDataAdapter import StiDataAdapter
from .classes.StiDataResult import StiDataResult
import pyodbc
from pyodbc import Connection

class StiOdbcAdapter(StiDataAdapter):
    version: str = '2023.4.1'
    checkVersion: bool = True
    connectionLink: Connection

    def connect(self):
        try:
            self.connectionLink = pyodbc.connect(self.connectionString)
            if (self.connectionInfo.charset):
                self.connectionLink.setdecoding(pyodbc.SQL_CHAR, self.connectionInfo.charset)
                self.connectionLink.setdecoding(pyodbc.SQL_WCHAR, self.connectionInfo.charset)
        except Exception as e:
            message = str(e)
            return StiDataResult.getError(self, message)
        
        return StiDataResult.getSuccess(self)
    
    def process(self):
        if (not super().process()):
            return False

        parameterNames = {
            'driver': ['driver'],
            'userId': ['uid', 'user', 'username', 'userid', 'user id'],
            'password': ['pwd', 'password']
        }

        return self.parseParameters(parameterNames)
    
    


class StiEvent:

    __handlers: list = None
    __name: str = None

    @property
    def count(self):
        return len(self.__handlers or '')
    
    @property
    def name(self):
        return self.__name
    
    @property
    def jsHandlers(self):
        return [handler for handler in self.__handlers if type(handler) == str or handler is True]
    
    def __iadd__(self, handler):
        self.__handlers.append(handler)
        return self
    
    def __isub__(self, handler):
        self.__handlers.remove(handler)
        return self

    def __call__(self, *args, **keywargs):
        for handler in self.__handlers:
            if (callable(handler)):
                handler(*args, **keywargs)

    def __init__(self, name: str):
        self.__handlers = []
        self.__name = name
	


from setuptools import setup, find_packages

with open('README.md') as file:
    readmeFile = file.read()

with open('REQUIREMENTS.txt') as file:
    requiresFile = [text.strip() for text in file if text.strip()]

setup(
    name = 'stimulsoft_data_adapters',
    version = '2023.4.1',
    author = 'Stimulsoft',
    author_email = 'info@stimulsoft.com',
    description = 'Stimulsoft Data Adapters for Python',
    long_description = readmeFile,
    long_description_content_type = 'text/markdown',
    url = 'https://www.stimulsoft.com/en',
    license = 'https://www.stimulsoft.com/en/licensing/developers',
    install_requires = ['pyodbc'],
    extras_require = {'ext': requiresFile},
    packages = find_packages(),
    python_requires = '>=3.7'
)



from .StiDataAdapter import StiDataAdapter
from .classes.StiDataResult import StiDataResult
import oracledb
from oracledb import FetchInfo

class StiOracleAdapter(StiDataAdapter):
    version: str = '2023.4.1'
    checkVersion: bool = True

    def getOdbcConnectionString(self):
        connectionString: str = \
            f'Driver={self.connectionInfo.driver};' \
            f'Dbq={self.connectionInfo.database};' \
            f'Uid={self.connectionInfo.userId};' \
            f'Pwd={self.connectionInfo.password};'
        
        return connectionString

    def connect(self):
        if (self.connectionInfo.driver):
            return self.connectOdbc()
        
        try:
            oracledb.init_oracle_client()
            self.connectionLink = oracledb.connect(
                user = self.connectionInfo.userId,
                password = self.connectionInfo.password,
                dsn = self.connectionInfo.database,
                encoding = self.connectionInfo.charset,
                mode = self.connectionInfo.privilege)
        except Exception as e:
            message = str(e)
            return StiDataResult.getError(self, message)
        
        return StiDataResult.getSuccess(self)
    
    def process(self):
        if (not super().process()):
            return False

        self.connectionInfo.port = 3306
        self.connectionInfo.privilege = oracledb.AUTH_MODE_DEFAULT

        parameterNames = {
            'driver': ['driver'],
            'database': ['database', 'data source', 'dbname'],
            'userId': ['uid', 'user', 'user id'],
            'password': ['pwd', 'password'],
            'charset': ['charset']
        }

        return self.parseParameters(parameterNames)
    
    def parseUnknownParameter(self, parameter: str, name: str, value: str):
        super().parseUnknownParameter(parameter, name, value)
        
        if (name.lower() == 'dba privilege' or name.lower() == 'privilege'):
            value = value.lower()
            if (value == 'sysoper' or value == 'oci_sysoper'):
                self.connectionInfo.privilege = oracledb.AUTH_MODE_SYSOPER
            if (value == 'sysdba' or value == 'oci_sysdba'):
                self.connectionInfo.privilege = oracledb.AUTH_MODE_SYSDBA

    def parseType(self, meta: FetchInfo):
        if (self.connectionInfo.driver):
            return super().parseType(meta)
        
        types = {
            'int': [oracledb.DB_TYPE_NUMBER],
            'number': [oracledb.DB_TYPE_BINARY_DOUBLE, oracledb.DB_TYPE_BINARY_FLOAT],
            'time': [oracledb.DB_TYPE_INTERVAL_DS],
            'datetime': [oracledb.DB_TYPE_DATE, oracledb.DB_TYPE_TIMESTAMP, oracledb.DB_TYPE_TIMESTAMP_LTZ, oracledb.DB_TYPE_TIMESTAMP_TZ],
            'blob': [oracledb.DB_TYPE_BFILE, oracledb.DB_TYPE_BLOB, oracledb.DB_TYPE_CLOB, oracledb.DB_TYPE_CURSOR, oracledb.DB_TYPE_NCLOB],
            'array': [oracledb.DB_TYPE_LONG_RAW, oracledb.DB_TYPE_OBJECT, oracledb.DB_TYPE_RAW],
            'string': [oracledb.DB_TYPE_CHAR, oracledb.DB_TYPE_JSON, oracledb.DB_TYPE_LONG, oracledb.DB_TYPE_NCHAR, oracledb.DB_TYPE_NVARCHAR,
                       oracledb.DB_TYPE_ROWID, oracledb.DB_TYPE_ROWID, oracledb.DB_TYPE_UROWID, oracledb.DB_TYPE_VARCHAR]
        }

        for key, array in types.items():
            if (meta.type in array):
                if (key == 'int'):
                    return 'number' if meta.scale > 0 else 'int'
                if (key == 'blob'):
                    return 'array'
                return key

        return 'string'
    
    def getValue(self, value: object, valueType: str):
        if (value is None):
            return None

        if (valueType == 'array'):
            if (type(value) == oracledb.LOB):
                value = value.read()

        return super().getValue(value, valueType)

    def makeQuery(self, procedure: str, parameters: list):
        paramsString = super().makeQuery(procedure, parameters)
        return f"SQLEXEC 'CALL {procedure} {paramsString}'"